from sqlalchemy import Table, Boolean, Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

Base = declarative_base()
session = None


class File(Base):
    __tablename__ = 'files'
    id = Column(Integer, primary_key=True)
    hash = Column(String, nullable=False)
    file = Column(String, nullable=False)
    packed = Column(Boolean, default=False, nullable=False)

    def add(hash: str, file: str) -> 'File':
        new_file = File(hash=hash, file=file)
        session.add(new_file)
        session.commit()
        return new_file


class Tag(Base):
    __tablename__ = 'tags'
    id = Column(Integer, primary_key=True)
    hash = Column(String, nullable=False)
    tag = Column(String, nullable=False)

    def add(hash: str, tag: str) -> 'Tag':
        new_tag = Tag(hash=hash, tag=tag)
        session.add(new_tag)
        session.commit()
        return new_tag


def bind(engine):
    global Base
    global session

    session = Session(engine)
    Base.metadata.create_all(engine)



import click


@click.group()
@click.pass_context
def cli(ctx):
    """ A command line utility for organizing directories of binary content.
    """
    ctx.ensure_object(dict)


@cli.command()
@click.pass_obj
def init(obj):
    """ Create a new fit repository in the current directory.
    """
    if not os.path.exists(obj['fit_path']):
        os.mkdir(obj['fit_path'])
        os.mkdir(os.path.join(obj['fit_path'], 'objects'))
        click.echo(f'Initialized empty fit repository in: {obj["fit_path"]!r}')
    else:
        click.echo(f'fit repository already exists in: {obj["fit_path"]!r}')


@cli.command()
@click.option('--clean', '-c', default=False, is_flag=True,
              help='Remove untracked files.')
@click.option('--force', '-f', default=False, is_flag=True,
              help='Don\'t prompt for confirmation when removing untracked '
              'files.')
@click.pass_obj
def pack(obj, clean, force):
    """ Removes all tracked links and their directories.
    """
    if not in_repo(obj):
        return

    unpacked_files = models.session.query(models.File).filter_by(packed=False)

    # Pack files
    for file in unpacked_files:
        click.echo(f'{file.hash} PACK {format_filename(obj, file.file)!r}')
        os.remove(file.file)
        file.packed = True
    models.session.commit()

    remove_empty_directories(obj['fit_repo_path'], clean=clean, force=force)


@cli.command()
@click.pass_obj
def status(obj):
    """ Show the active directory status.
    """
    if not in_repo(obj):
        return

    file_query = models.session.query(models.File)
    packed_files = file_query.filter_by(packed=True)
    tag_query = models.session.query(models.Tag).group_by(models.Tag.tag)
    click.echo(f'{file_query.count()} tracked files - {packed_files.count()} packed')
    click.echo(f'{tag_query.count()} tags')


@cli.group()
@click.pass_context
def tag(ctx):
    """ Create, manage, and view tags.
    """
    if not in_repo(obj):
        return

    ctx.ensure_object(dict)


@tag.command()
@click.argument('tag')
@click.argument('filenames', type=click.Path(exists=True), nargs=-1,
                required=True)
@click.pass_obj
def add(obj, tag, filenames):
    """ Tag one or more files.
    """
    if not in_repo(obj):
        return

    for filename in filenames:
        file = models.session.query(models.File) \
                .filter_by(file=os.path.abspath(filename)) \
                .first()
        if file:
            hash = file.hash
            tag_query = models.session.query(models.Tag) \
                .filter_by(hash=hash, tag=tag)
            if not tag_query.count():
                models.Tag.add(hash=hash, tag=tag)
                click.echo(f'{hash} TAGGED {tag!r}')
            else:
                click.echo(f'{hash} is already tagged {tag!r}')
        else:
            click.echo(f'Cannot tag untracked file: {format_filename(obj, filename)!r}')


@tag.command()
@click.argument('tag')
@click.argument('filenames', type=click.Path(exists=True), nargs=-1,
                required=True)
@click.pass_obj
def remove(obj, tag, filenames):
    """ Remove tag from one or more files.
    """
    if not in_repo(obj):
        return

    for filename in filenames:
        file = models.session.query(models.File) \
                .filter_by(file=os.path.abspath(filename)) \
                .first()
        if file:
            hash = file.hash
            tag_query = models.session.query(models.Tag) \
                .filter_by(hash=hash, tag=tag)
            if tag_query.count():
                tag_query.delete()
                models.session.commit()
                click.echo(f'{hash} UNTAGGED {tag!r}')
            else:
                click.echo(f'{hash} is not tagged {tag!r}')
        else:
            click.echo(f'Cannot tag untracked file: {format_filename(obj, filename)!r}')


@tag.command()
@click.argument('tag')
@click.pass_obj
def find(obj, tag):
    """ Find all files with a certain tag.
    """
    if not in_repo(obj):
        return

    tag_obj = models.session.query(models.Tag).filter_by(tag=tag).first()
    if tag_obj:
        file_query = models.session.query(models.File) \
            .filter_by(hash=tag_obj.hash) \
            .order_by(models.File.file)
        for file in file_query:
            click.echo(format_filename(obj, file.file))
    else:
        click.echo(f'No such tag: {tag!r}')


@tag.command()
@click.option('--uses', '-u', default=False, is_flag=True,
              help='Sort by and display number of uses.')
@click.pass_obj
def show(obj, uses):
    """ Show all existing tags.
    """
    if not in_repo(obj):
        return

    tag_query = models.session.query(
        models.Tag,
        func.count(models.Tag.tag).label('count')
    ) \
        .group_by(models.Tag.tag) \
        .order_by(desc('count') if uses else models.Tag.tag)
    for tag, count in tag_query:
        if uses:
            click.echo(f'{tag.tag} ({count} uses)')
        else:
            click.echo(tag.tag)


@tag.command()
@click.argument('filenames', type=click.Path(exists=True), nargs=-1,
                required=True)
@click.pass_obj
def view(obj, filenames):
    """ View all the tags for one or more files.
    """
    if not in_repo(obj):
        return

    for filename in filenames:
        file = models.session.query(models.File) \
                .filter_by(file=os.path.abspath(filename)) \
                .first()
        if file:
            hash = file.hash
            tag_query = models.session.query(models.Tag) \
                .filter_by(hash=hash) \
                .order_by(models.Tag.tag)
            if tag_query.count():
                tags = [tag.tag.__repr__() for tag in tag_query]
                click.echo(f'{hash} {", ".join(tags)}')
        else:
            click.echo(f'Cannot tag untracked file: {format_filename(obj, filename)!r}')


@cli.command()
@click.option('--soft/--hard', '-s/-h', default=False,
              help='The type of links to create.')
@click.pass_obj
def unpack(obj, soft):
    """ Restores all tracked links and their directories.
    """
    if not in_repo(obj):
        return

    object_dir_path = os.path.join(obj['fit_path'], 'objects')
    packed_files = models.session.query(models.File).filter_by(packed=True)
    for file in packed_files:
        object_path = os.path.join(object_dir_path, file.hash)
        file_path = os.path.relpath(file.file, obj['fit_repo_path'])
        for dir_part in [os.path.sep.join(file_path.split(os.path.sep)[:i+1])
                         for i in range(len(file_path.split(os.path.sep)) - 1)]:
            dir_part = os.path.join(obj['fit_repo_path'], dir_part)
            if not os.path.exists(dir_part):
                os.mkdir(dir_part)

        # Create links
        click.echo(f'{file.hash} LINK ({"soft" if soft else "hard"}) {format_filename(obj, file.file)!r}')
        if soft:
            os.symlink(object_path, file.file)
        else:
            os.link(object_path, file.file)

        file.packed = False
    models.session.commit()


@cli.command()
@click.option('--soft/--hard', '-s/-h', default=False,
              help='The type of links to create.')
@click.option('--pack', '-p', default=False, is_flag=True,
              help='Automatically pack new files.')
@click.pass_obj
def update(obj, soft, pack):
    """ Tracks changes to the active directory and replaces literal files with
        their respective links.
    """
    if not in_repo(obj):
        return

    # Make list of tracked files that are expected to be found
    expected_file_query = models.session.query(models.File.file) \
            .filter_by(packed=False)
    expected_files = [row.file for row in expected_file_query]

    # Traverse active directory looking for changes
    for root, dirs, files in os.walk(obj['fit_repo_path'], topdown=True):
        # Filter excluded paths
        dirs[:] = [d for d in dirs if not exclude(d)]
        # Process files
        for file in files:
            file_path = os.path.join(root, file)
            if file_path in expected_files:
                expected_files.remove(file_path)
            if not exclude(file):
                process_file(obj, file_path, soft=soft)

    remove_empty_directories(obj['fit_repo_path'])

    # Remove unpacked file rows that were not found in active directory
    for file_path in expected_files:
        click.echo(f'{format_filename(obj, file_path)!r} REMOVE (outdated file reference)')
        file = models.session.query(models.File) \
                .filter_by(file=file_path).delete()
        models.session.commit()

    # Remove newly orphaned objects
    for hash in os.listdir(os.path.join(obj['fit_path'], 'objects')):
        file_query = models.session.query(models.File).filter_by(hash=hash)
        if not file_query.count():
            click.echo(f'{hash} REMOVE (orphaned object)')
            object = os.path.join(obj['fit_path'], 'objects', hash)
            os.remove(object)


if __name__ == '__main__':
    obj = dict()
    obj['fit_path'] = os.path.abspath(find_path_up('.fit', start='.') or './.fit')
    obj['db_path'] = os.path.join(obj['fit_path'], 'data')
    obj['fit_repo_path'] = os.path.abspath(os.path.join(obj['fit_path'], '..'))
    if os.path.exists(obj['fit_path']):
        obj['db'] = create_engine(f'sqlite:///{obj["db_path"]}')
    else:
        obj['db'] = create_engine(f'sqlite:///:memory:')
    models.bind(obj['db'])
    obj['in_repo'] = os.path.exists(obj['fit_path'])
    cli(obj=obj)



from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name='fit-cli',
    version='0.1.0',
    description='A command line utility for organizing directories of binary content.',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Jake Ledoux',
    author_email='contactjakeledoux@gmail.com',
    url='https://github.com/jakeledoux/fit-cli',
    license='GNU General Public License v2.0',
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'click>=7.1.2',
        'imageio>=2.9.0',
        'numpy>=1.20.1',
        'Pillow>=8.1.0',
        'SQLAlchemy>=1.3.23'
    ],
    entry_points='''
        [console_scripts]
        fit=fit.scripts.fit:cli
    '''
)



""" fit
"""
import click
import hashlib
import imageio
import json
from . import models
import os
import shutil
from sqlalchemy import create_engine, desc, func
import sys
from typing import List, Optional


def exclude(name: str) -> bool:
    """ Checks if file or directory with `name` should be excluded from
        `os.walk`.
    """
    return name.startswith('.')


def find_path_up(filename: str, start: str = '.') -> Optional[str]:
    """ Searchs for `filename` in `start` and all its parent directories until
        `filename` is found or top level is reached.
    """
    start = os.path.abspath(start)
    for dir_part in [os.path.sep.join(start.split(os.path.sep)[:i+1])
                     for i in reversed(
                         range(len(start.split(os.path.sep)))
                     )]:
        if not dir_part:
            dir_part = '/'
        if filename in os.listdir(dir_part):
            return os.path.join(dir_part, filename)


def format_filename(obj, filename: str) -> str:
    """ Formats path for display by making it relative to the repo dir.
    """
    rel_path = os.path.relpath(filename, obj['fit_repo_path'])
    return click.format_filename(rel_path)


def get_keywords(file_path: str) -> List[str]:
    """ Gets XP keywords from file contents.
    """
    try:
        with open(file_path, 'rb') as f:
            image = Image(f.read())
        return image.get('xp_keywords', '').split(';')
    except:
        return []


def hash_file(file_path: str) -> str:
    """ Gets SHA-1 hash of file contents (or pixels if file is an image).
    """
    # Attempt to load image
    try:
        file_bytes = imageio.imread(file_path)
        if not file_bytes.flags.c_contiguous:
            file_bytes = file_bytes.copy(order='C')
    # Load raw bytes of file
    except:
        with open(file_path, 'rb') as f:
            file_bytes = f.read()
    return hashlib.sha1(file_bytes).hexdigest()


def in_repo(obj) -> bool:
    """ Check if current context is within a fit repository.
    """
    if obj['in_repo']:
        return True
    click.echo('fatal: not a fit repository')
    return False


def is_orphan(file_path: str) -> bool:
    """ Checks if number of inode references is one.
    """
    return os.stat(file_path).st_nlink < 2


def process_file(obj, file_path: str, soft=False):
    """ Tracks file in fit repo.
    """
    file_ext = os.path.splitext(file_path)[1]

    # If file is not hard link
    file_query = models.session.query(models.File).filter_by(file=file_path)
    if not file_query.count():
        hash = hash_file(file_path)

        # Append filename to filenames list
        file = models.File.add(hash=hash, file=file_path)

        # Create object if necessary and remove file
        object = os.path.join(obj['fit_path'], 'objects', hash)
        if os.path.exists(object):
            os.remove(file_path)
        else:
            click.echo(f'{hash} CREATE from {format_filename(obj, file_path)!r}')
            os.rename(file_path, object)

        if pack:
            file.packed = True
            models.session.commit()
        else:
            # Create links
            click.echo(f'{hash} LINK ({"soft" if soft else "hard"}) {format_filename(obj, file_path)!r}')
            if soft:
                os.symlink(object, file_path)
            else:
                os.link(object, file_path)


def remove_empty_directories(path: str, clean=False, force=False):
    """ Removes all empty directories in a given path.
    """
    for file in os.listdir(obj['fit_repo_path']):
        if os.path.isdir(file) and not exclude(file):
            for root, dirs, files in os.walk(file, topdown=False):
                if clean:
                    for file in files:
                        file_path = os.path.join(root, file)
                        click.echo(f'Found untracked file: {format_filename(obj, file_path)!r}')
                        if force or click.confirm('Remove?'):
                            click.echo(f'{format_filename(obj, file_path)!r} REMOVE')
                            os.remove(file_path)
                if not os.listdir(root):
                    file_path = os.path.join(obj['fit_repo_path'], root)
                    click.echo(f'{format_filename(obj, file_path)!r} REMOVE (empty directory)')
                    os.rmdir(file_path)
